;;; boot.scm - Dezyne ide Guile startup

(use-modules (srfi srfi-1)
             (srfi srfi-26)
             (ice-9 regex)
             (ice-9 pretty-print)
             (json)
             (web socket server)

             (ide wx bindings)
             (ide console)
             (ide command)
             (ide completion)
             (ide dzn)
             (ide edit)
             (ide help)
             (ide interface)
             (ide introspection)
             (ide minibuffer)
             (ide keyboard)
             (ide util)
             (ide window))

(define (get-url what)
  (cond ((equal? what "state") (string-append "file://" (getenv "NPMLIB") "/node_modules/dzn-enterprise/view_state.html"))
        ((equal? what "system") (string-append "file://" (getenv "NPMLIB") "/node_modules/dzn-enterprise/view_system.html"))
        ((equal? what "trace") (string-append "file://" (getenv "NPMLIB") "/node_modules/dzn-enterprise/view_trace.html"))
        (else "")))

(define %data-prefix (let ((here (dirname (current-filename))))
                       (if (access? (string-append here "/icons") F_OK) here
                           (string-append here "/../../../dezyne-ide"))))

(set! %prefix %data-prefix)

(define c-commands
  `((auto-complete . ,auto-complete)
    (linum-mode . ,linum-mode)
    (mark-whole-buffer . ,mark-whole-buffer)
    (menu-bar-mode . ,menu-bar-mode)
    (recenter . ,recenter)
    (set-mark . ,set-mark)
    (tool-bar-mode . ,tool-bar-mode)))

(for-each set-command-properties! (map cdr c-commands))

(define c-functions
  `((about . ,about)
    (buffer-file-name . ,buffer-file-name)
    (buffer-substring . ,buffer-substring)
    (frame-set-size . ,frame-set-size)
    (frame-full-screen . ,frame-full-screen)
    (key-command . ,key-command)
    (line-len . ,line-len)
    (point . ,point)
    (point-max . ,point-max)
    (anchor . ,anchor)
    (show-location . ,show-location)
    (move-point . ,move-point)
    (unselect . ,unselect)
    (visit-file . ,visit-file)
    (visit-url . ,visit-url)
    (minibuffer-active? . ,minibuffer-active?)
    (minibuffer-clear . ,minibuffer-clear)
    (minibuffer-insert . ,minibuffer-insert)
    (minibuffer-return-focus . ,minibuffer-return-focus)
    (minibuffer-prefix . ,minibuffer-prefix)
    (minibuffer-quit . ,minibuffer-quit)
    (minibuffer-string . ,minibuffer-string)
    (console-pipe-open . ,console-pipe-open)
    (console-delete-process . ,console-delete-process)
    (console-pipe-process-terminated? . ,console-pipe-process-terminated?)
    (console-pipe-read . ,console-pipe-read)
    (console-pipe-can-read? . ,console-pipe-can-read?)
    (console-pipe-error-read . ,console-pipe-error-read)
    (console-pipe-can-error-read? . ,console-pipe-can-error-read?)
    (console-pipe-process-interrupt . ,console-pipe-process-interrupt)
    (console-pipe-process-terminate . ,console-pipe-process-terminate)
    (console-activate . ,console-activate)
    (console-active? . ,console-active?)
    (console-clear . ,console-clear)
    (console-insert . ,console-insert)
    (console-string . ,console-string)
    (console-start-idle-hook . ,console-start-idle-hook)
    (console-stop-idle-hook . ,console-stop-idle-hook)
    (linum-mode . ,linum-mode)
    (message . ,message)
    (indent . ,indent)
    (insert . ,insert)

    (application-exit . ,application-exit)
    (marked . ,marked)
    (reload . ,reload)

    (find-file-dialog . ,find-file-dialog)

    (create-tool-bar-separator . ,create-tool-bar-separator)
    (create-tool-bar-button . ,create-tool-bar-button)
    (create-menu-entry . ,create-menu-entry)

    (help-visible? . ,help-visible?)
    (help-toggle . ,help-toggle)
    (message-box . ,message-box)))

(let ((module (resolve-module '(ide interface)))
      (all (append c-commands c-functions)))
  (for-each
   (lambda (x)
     (let ((name (car x))
           (value (cdr x)))
       (module-define! module name value)))
   all)
  (module-export! module (map car all)))

;;; MENU
(define (file-new)
  (visit-file ""))

(define (make-non-blocking socket)
  (if (mingw?)
      (make-non-blocking-mingw socket)
      (fcntl socket F_SETFL (logior O_NONBLOCK (fcntl socket F_GETFL)))))

(define (listen-for-location)
  (let ((socket (make-server-socket #:port 4321)))
    (listen socket 1)
    (unless (mingw?) (sigaction SIGPIPE SIG_IGN))
    (make-non-blocking socket)
    socket))

(define* (handle-location server-socket)
  (let ((socket (and=> (and (port? server-socket)
                            (not (port-closed? server-socket))
                            (accept server-socket)) car)))
    (when socket
      ((@@ (web socket server) serve-client) socket
       (lambda (data)
         (select-location data)
         "ack"))
      (close-port socket))))

(define (select-location data)
  (unless (string-null? data)
   (let* ((location (hash-map->list cons (json-string->scm data)))
          (file-name (assoc-ref location "file-name"))
          (start (string->number (assoc-ref location "offset")))
          (end (+ (string->number (assoc-ref location "length")) start)))
     (visit-file file-name)
     (move-point end)
     (anchor start)
     (recenter))))

(define (handle id)
  (case id
    ((200) (file-new))
    ((201) (find-file-dialog))
    ((203) (reload))
    ;;((204) (kill-buffer))
    ((205) (application-exit))
    ((206) (key-command STC-CUT))
    ((207) (key-command STC-COPY))
    ((208) (key-command STC-PASTE))
    ((209) (key-command STC-UNDO))
    ((210) (key-command STC-REDO))
    ((212) (dzn-explore))
    ((213) (info))
    ((214) (about))
    (else  (pretty-print (format #t "not handled: ~a" id)))))

(define menu
  '(("File"  ((200 "New" "icons/gtk-new.png" "New file")
              (201 "Open" "icons/gtk-open.png" "Open file")
              (202 "Save" "icons/gtk-save.png" "Save file" 'disabled)
              ;;(204 "Close" "icons/gtk-close.png" "Close file")
              (205 "Quit" "icons/gtk-quit.png" "Quit program")))
    ("Edit"  ((206 "Cut" "icons/gtk-cut.png" "Cut selection")
              (207 "Copy" "icons/gtk-copy.png" "Copy selection")
              (208 "Paste" "icons/gtk-paste.png" "Paste cut/copy selection")
              (209 "Undo" "icons/gtk-undo-ltr.png" "Undo last modification")
              (210 "Redo" "icons/gtk-redo-ltr.png" "Redo last modification")
              (211 "Search" "icons/gtk-find.png" "Search buffer")))
    ("Model" ((212 "Explore" "icons/gtk-find-and-replace.png" "Explore model")
              (203 "Reload" "icons/gtk-refresh.png" "Reload view")))
    ("Help"  ((213 "Documentation" "icons/gtk-help.png" "Open documentation")
              (214 "About" "icons/gtk-about.png" "Show about")))))

(define (prefix-menu-entry me)
  (list (first me) (second me) (string-append %data-prefix "/" (third me)) (fourth me)))

;;; SETUP
(define (setup)
  (frame-set-size)
  (for-each (lambda (m)
              (for-each (lambda (e)
                          (create-menu-entry (first e) (car m) (second e))
                          (apply create-tool-bar-button (prefix-menu-entry e)))
                        (cadr m))
              (create-tool-bar-separator))
            menu)
  (console-reset))

(define (load-dezyne.scm)
  (let* ((dezyne.scm (string-append (getenv "HOME") "/.dezyne.scm"))
         (init (and (access? dezyne.scm R_OK) dezyne.scm)))
    (and init (load init))
    (after-cua-mode-hook)))

(define (exception-safe-call lst)
  (catch #t
   (lambda _ (apply (car lst) (cdr lst)))
   (lambda (key . args)
     (format (current-error-port) "caught exception key: ~a args: ~s\n" key args)
     (message-box (format #f "caught exception: ~a: ~s\n" key args)))
   (lambda _ (backtrace))))

;;; KEYBOARD
(define (key-press key mod)
  (when (getenv "DEBUG_IDE")
    (format (current-error-port) "\nkey-press key=~s mod=~s => code=~s\n" key mod (keycode key mod)))
  (let ((binding (and=> (assoc (keycode key mod) current-key-map (lambda (k c) (eq? k (kbd c)))) cdr)))
    (cond ((procedure? binding)
	   (set! current-key-map global-key-map)
	   (binding)
	   #t)
	  ((pair? binding)
	   (set! current-key-map binding)
	   #t)
	  (else
	   (cond ((console-active?) (set! current-key-map console-key-map))
                 ((minibuffer-active?) (set! current-key-map minibuffer-key-map))
                 (else (set! current-key-map global-key-map)))
	   #f))))

(set! ctrl-h-map
  `(("b" . ,help-bindings)
    ("e" . ,help-explore)
    ("f" . ,describe-function)
    ("F" . ,describe-command)
    ("i" . ,info)
    ("h" . ,help)
    ("k" . ,describe-key)
    ("?" . ,help)))

(set! ctrl-x-map
  `(("C-g" . ,global-quit)
    ("C-c" . ,application-exit)
    ("C-f" . ,find-file)
    ("C-o" . ,find-file-dialog)
    ("=" . ,what-cursor-position)
    ("h" . ,mark-whole-buffer)))

(define scintilla-edit-map ; Use string to describe/document scintilla commands
  `(("C-c" . "copy region")
    ("C-v" . "yank")
    ("C-x" . "delete region")
    ("C-z" . "undo")))

(set! emacs-global-key-map
  `(("<left>" . ,backward-char)
    ("<up>" . ,previous-line)
    ("<right>" . ,forward-char)
    ("<down>" . ,next-line)
    ("C-g" . ,kill-minibuffer-or-quit)

    ;; FIXME
    ;;("C-x" . ,(lambda _ (message "C-x-") ctrl-x-map))
    ;;("C-h" . ,(lambda _ (message "C-h (type ? for further options)") ctrl-h-map))

    ("C-x" . ,ctrl-x-map)
    ("C-h" . ,ctrl-h-map)

    ("C-k" . ,kill-line)
    ("C-l" . ,recenter)
    ("C-y" . ,yank)
    ("C-b" . ,backward-char)
    ("C-f" . ,forward-char)
    ("M-b" . ,backward-word)
    ("M-f" . ,forward-word)
    ("C-p" . ,previous-line)
    ("C-n" . ,next-line)
    ("C-a" . ,beginning-of-line)
    ("C-e" . ,end-of-line)
    ("M-<" . ,beginning-of-buffer)
    ("M-." . ,end-of-buffer)
    ;;("C-M-n" . ,(lambda _ (dzn-find-previous-open-brace)))
    ;;("C-M-p" . ,(lambda _ (dzn-find-next-closing-brace)))
    ("C-d" . ,delete-char)
    ("M-d" . ,kill-word)
    ("M-DEL" . ,backward-kill-word)
    ("C--" . ,zoom-out)
    ("C-+" . ,zoom-in)
    ("C-=" . ,zoom-in)
    ("C-_" . ,undo)
    ("C-/" . ,undo)
    ("C-o" . ,open-line)
    ("C-w" . ,kill-region)
    ("M-w" . ,copy-region)
    ("M-u" . ,upcase-word)
    ("M-l" . ,downcase-word)
    ("C-s" . ,isearch-forward)
    ("M-/" . ,auto-complete)
    ("C-SPACE" . ,set-mark)
    ;;("M-\\" . ,deindent)
    ;;("TAB" . ,indent)
    ("M-x" . ,execute-extended-command)
    ("M-:" . ,eval-expression)
    ("M-!" . ,shell-command)
    ("M-&" . ,async-shell-command)
    ("ESC" . ,help-bury)
    ("<f1>" . ,info)
    ("<f5>" . ,reload)
    ("<f11>" . ,frame-full-screen)
    ("C-r" . ,reload)
    ,@scintilla-edit-map))

(set! windows-global-key-map
  `(("C-SPACE" . ,auto-complete)
    ("M-<f4>" . ,exit)
    ("M-x" . ,execute-extended-command)
    ("M-:" . ,eval-expression)
    ("ESC" . ,help-bury)
    ("<f1>" . ,info)
    ("<f5>" . ,reload)
    ("<f11>" . ,frame-full-screen)
    ,@scintilla-edit-map))

;; Enable cua mode by default, expect Emacs users to be willing and
;; able to cp emacs-init.scm ~/.dezyne.scm
(set! global-key-map windows-global-key-map)

(define (string-replace-string from to string)
  (cond ((string-contains string from) => (lambda (i) (string-replace string to i (+ i (string-length from)))))
        (else string)))

(define (html-escape x)
  (let ((escapes '(("&" . "&amp;")
                   ("<" . "&lt;")
                   (">" . "&gt;")
                   ("\"" . "&quot;"))))
    (fold string-replace-string x (map car escapes) (map cdr escapes))))

(define (kbd-entry->string entry)
  (let ((proc (cdr entry)))
    (cond ((string? proc) (format #f "~a\t\t~a" (car entry) proc))
          ((procedure? proc)
           (let ((key (html-escape (format #f "~a" (car entry))))
                 (name (html-escape (format #f "~a" (or (procedure-name proc) "<lambda>")))))
                (format #f "~a\t\t<a href=\"command/~a.html\">~a</a>" key name name)))
          (else #f))))

(use-modules (ide minibuffer))
(when (getenv "DUMP_DOC")
  (system "mkdir -p info/command")
  (display (string-join (filter-map kbd-entry->string emacs-global-key-map) "\n"))
  (with-output-to-file "info/emacs-keys.html"
    (lambda _
      (let ((entries (append (filter-map kbd-entry->string emacs-global-key-map)
                             (map (cut string-append "C-h " <>)
                                  (filter-map kbd-entry->string ctrl-h-map))
                             (map (cut string-append "C-x " <>)
                                  (filter-map kbd-entry->string ctrl-x-map)))))
        (display "<pre>\n")
        (display (string-join (sort entries string<?) "\n"))
        (newline)
        (display "</pre>\n"))))
  (with-output-to-file "info/windows-keys.html"
    (lambda _
      (let ((entries (append (filter-map kbd-entry->string windows-global-key-map))))
        (display "<pre>\n")
        (display (string-join (sort entries string<?) "\n"))
        (newline)
        (display "</pre>\n"))))
  (for-each
   (lambda (c)
     (let* ((name (car c))
            (proc ((compose safe-variable-ref cdr) c))
            (properties (source-properties proc))
            (location (or (program-location proc)
                          (assoc-ref properties 'location)))
            (file-name (or (assoc-ref location "file-name")
                           "unknown"))
            (prefix (dirname (current-filename)))
            (relative (if (string-prefix? prefix file-name) (string-append "../../" (substring file-name (string-length prefix)))
                          (string-append "../../" file-name)))
            (colon-line (or (and=> (assoc-ref location "line")
                                   (cut format #f ":~a" <>))
                            ""))
            (hash-line (or (and=> (assoc-ref location "line")
                                  (cut format #f "#~a" <>))
                           ""))
            (documentation (or (procedure-documentation proc)
                               (assoc-ref properties "documentation")
                               "Not documented."))
            (cls (if (command? proc) "Command" "Procedure"))
            (type (if (string-suffix? ".scm" file-name) "Scheme function"
                      "C function"))
            (content (string-append "<pre>\n"
                                    (format #f "~a ~a is a ~a defined in <a href=\"~a~a\">[~a~a]</a>\n\n" cls name type relative hash-line file-name colon-line)
                                    documentation
                                    "</pre>\n")))
       (with-output-to-file (format #f "info/command/~a.html" name)
         (lambda _
           (display content)))))
   (delete-duplicates (append (all-commands) (all-ide-procedures)) eq?))
  (application-exit))

;;; completion/parser test
;; (let ((data ((compose read-string open-input-file) "test.dzn")))
;;     (for-each (lambda (i)
;;              (warn '=> (completions data i)))
;;            (iota (/ (string-length data) 2) 0 2)))
