(define-module (ide console)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-2)
  #:use-module (srfi srfi-26)

  #:use-module (ice-9 popen)
  #:use-module (ice-9 rdelim)
  #:use-module (ice-9 receive)

  #:use-module (system foreign)

  #:use-module (ide command)
  #:use-module (ide edit)
  #:use-module (ide interface)
  #:use-module (ide keyboard)
  #:use-module (ide util)

  #:export (console-clear-line
            console-interrupt
            console-reset
            console-return
            console-idle-hook
            run-shell-command
            run-shell-command-async
            %prompt))

(define %prompt "$ ")

(define (console-return)
  (let* ((string (console-string))
         (lines (string-split string #\newline))
         (line (string-trim-both (last lines)))
         (command (substring line (min (string-length line) (string-length %prompt)))))
    (end-of-buffer)
    (console-insert "\n")
    (if (string-null? command)
        (console-insert %prompt)
        (run-shell-command command))))

(define (console-clear-line)
  (let* ((string (console-string))
         (lines (string-split string #\newline)))
    (console-clear)
    (console-insert (string-join (drop-right lines 1)))
    (console-insert %prompt)))

(define (console-reset)
  (console-clear)
  (console-insert %prompt))

(define console-process #f)

(define (cleanup-previous-process thunk)
  (console-pipe-process-terminate
   console-process
   (lambda _
     (console-delete-process console-process)
     (console-stop-idle-hook)
     (set! console-idle-hooks '())
     (set! console-process #f)
     (console-insert %prompt)
     (thunk))))

(define (run-shell-command command)
  (if console-process
      (let ((status 0))
        (cleanup-previous-process (cut run-shell-command command))
        status)
      (receive (process status)
          (console-pipe-open (if (mingw?) (string-append "cmd /c " command)
                                 command))
        (when process
          (let ((stderr-port (make-soft-port
                              (vector (const #f) (const #f) (const #f)
                                      (cut console-pipe-error-read process)
                                      (const #f)
                                      (cut console-pipe-can-error-read? process))
                              "r"))
                (stdout-port (make-soft-port
                              (vector (const #f) (const #f) (const #f)
                                      (cut console-pipe-read process)
                                      (const #f)
                                      (cut console-pipe-can-read? process))
                              "r"))
                (port (make-soft-port
                       (vector (const #f) (cut console-insert <>) #f
                               (const #f) #f)
                       "w")))
            (display (read-string stderr-port) port)
            (display (read-string stdout-port) port)
            (console-delete-process process)
            (console-insert %prompt)))
        status)))

(define console-idle-hooks '())

(define (console-idle-hook)
  (for-each (cut apply <> '()) console-idle-hooks))

(define (console-add-idle-hook hook)
  (set! console-idle-hooks (cons hook console-idle-hooks))
  (console-start-idle-hook))

(define (console-remove-idle-hook hook)
  (set! console-idle-hooks (filter (negate (cut eq? hook <>)) console-idle-hooks))
  (when (null? console-idle-hooks)
    (console-stop-idle-hook)))

(define-interactive (console-interrupt)
  "Kill running console process."
  (when console-process
    (console-pipe-process-interrupt console-process)
    (console-insert "^C\n")))

(define (run-shell-command-async command)
  (if console-process
      (cleanup-previous-process (cut run-shell-command-async command))
      (begin
        (when (string-suffix? %prompt (console-string))
          (console-insert command)
          (console-insert "\n"))
        (let ((process (console-pipe-open (if (mingw?) (string-append "cmd /c " command)
                                              command) 'async)))
          (when process
            (let* ((input-port (make-soft-port
                                (vector (const #f) (const #f) (const #f)
                                        (cut console-pipe-read process)
                                        (const #f)
                                        (cut console-pipe-can-read? process))
                                "r"))
                   (error-port (make-soft-port
                                (vector (const #f) (const #f) (const #f)
                                        (cut console-pipe-error-read process)
                                        (const #f)
                                        (cut console-pipe-can-error-read? process))
                                "r"))
                   (output-port (make-soft-port
                                 (vector (const #f) (cut console-insert <>) #f
                                         (const #f) #f)
                                 "w")))
              (define (hook)
                (catch #t
                       (lambda _
                         (while (or (char-ready? error-port)
                                    (char-ready? input-port))
                                (if (char-ready? error-port)
                                    (let ((line (read-line error-port)))
                                      (when (not (eof-object? line))
                                        (when (string-prefix? (string-append (warn 'bfn (buffer-file-name)) ":") line)
                                          (format (current-error-port) "error!"))
                                        (display line output-port)
                                        (newline output-port))))
                                (if (char-ready? input-port)
                                    (let ((line (read-line input-port)))
                                      (when (not (eof-object? line))
                                        (when (string=? line "initialisation done")
                                          (reload))
                                        (display line output-port)
                                        (newline output-port)))))

                         (when (console-pipe-process-terminated? process)
                           (console-delete-process process)
                           (console-remove-idle-hook hook)
                           (console-stop-idle-hook)
                           (set! console-process #f)
                           (console-insert %prompt)))

                       (lambda (key . args)
                         (format (current-error-port) "hook: catch ~s ~s" key args)
                         (console-remove-idle-hook hook)
                         (console-stop-idle-hook)
                         (console-pipe-process-terminate process)
                         (console-delete-process process)
                         (set! console-process #f)
                         (console-insert %prompt))))

              (set! console-process process)
              (console-add-idle-hook hook)))))))

(define-interactive (shell)
  (console-activate))
