(define-module (ide introspection)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (ice-9 session)
  #:use-module (system vm program)

  #:use-module (ide command)
  #:export (all-commands
            all-ide-procedures
            all-procedures
            all-variables
            program-location
            safe-variable-ref))

(define (module-variables module)
  (module-map cons module))

(define* (all-variables #:optional (module (current-module)))
  (append-map module-variables ((apropos-fold-accessible module) cons '())))

(define (safe-variable-ref o)
  (catch #t (lambda _ (variable-ref o)) (const #f)))

(define* (all-procedures #:optional (module (current-module)))
  (filter (compose procedure? safe-variable-ref cdr) (all-variables module)))

(define (procedure-documentation proc)
  (or ((@@ (guile-user) procedure-documentation) proc)
      (assoc-ref (source-properties proc) "documentation")))

(define (procedure-file-name proc)
  (let ((location (or (program-location proc)
                      (assoc-ref (source-properties proc) 'location))))
    (or (assoc-ref location "file-name")
        "unknown")))

(define ide-files '("dezyne.cc"
                    "command.scm"
                    "completion.scm"
                    "console.scm"
                    "dzn.scm"
                    "edit.scm"
                    "help.scm"
                    "interface.scm"
                    "introspection.scm"
                    "keyboard.scm"
                    "minibuffer.scm"
                    "text.scm"
                    "window.scm"))

(define* (all-ide-procedures #:optional (module (current-module)))
  (filter (compose (cut member <> ide-files) basename procedure-file-name safe-variable-ref cdr)
          (all-procedures module)))

(define* (all-commands #:optional (module (current-module)))
  (filter (compose command? safe-variable-ref cdr) (all-procedures module)))

(define (make-location file-name line)
  (list (cons "file-name" (and (string? file-name) file-name))
        (cons "file" (and (string? file-name) file-name)) ; compat
        (cons "line" (and (number? line) (+ 1 line)))))

(define (fill-file->module-mapping! ret)
  (define (record-module m)
    (let ((f (module-filename m)))
      (hash-set! ret f (cons m (hash-ref ret f '())))))
  (define (visit-module m)
    (record-module m)
    (hash-for-each (lambda (k v) (visit-module v))
                   (module-submodules m)))
  (visit-module (resolve-module '() #f))
  ret)

(define file->modules (fill-file->module-mapping! (make-hash-table)))

(define (program-file p)
  (let ((src (program-source p 0)))
    (and (pair? src) (cadr src))))

(define (program-module p)
  (let* ((f (program-file p))
         (mods (or (hash-ref file->modules f)
                   (hash-ref (fill-file->module-mapping! file->modules) f))))
    (and (pair? mods) (not (null? mods)) (car mods))))

(define (module-name? module-name)
  (and (list? module-name)
       (not (null? module-name))
       (every symbol? module-name)))

(define (module-file-name module-name)
  (and (module-name? module-name)
       (or ((@@ (ice-9 session) module-filename) module-name)
           (module-filename (resolve-module module-name #f)))))

(define (program-file-name p)
  (let* ((mod (program-module p))
         (name (and (module? mod) (module-name mod))))
    (and name (module-file-name name))))

(define (program-location p)
  (cond ((not (program? p)) #f)
        ((program-source p 0) =>
         (lambda (s) (make-location (program-file-name p) (source:line s))))
        ((program-file-name p) => (lambda (s) (make-location s #f)))
        (else #f)))
